/* Nessus
 * Copyright (c) mjh-EDV Beratung, 1999-2k
 * Copyright (C) 2k Renaud Deraison
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Author: Jordan Hrycaj <jordan@mjh.teddy-net.com>
 */

#include "includes.h"

#include <stdarg.h>

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_SIGNAL_H
#include <signal.h>
#endif

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

#ifdef HAVE_SELECT
/* not both, HAVE_SELECT and HAVE_POLL */
# undef HAVE_POLL
#else
# ifdef HAVE_POLL
#  ifdef HAVE_SYS_POLL_H
#   include <sys/poll.h>
#  endif
#  ifdef HAVE_POLL_H
#   include <poll.h>
#  endif
# else /* HAVE_POLL */
#  error You loose, neither select() not poll () avilable
# endif /* HAVE_POLL */
#endif /* HAVE_SELECT */

#if 0
#define HAVE_POLL
#define USE_PTHREADS
#define HAVE_PTHREAD_H
#include <sys/poll.h>
#endif

#ifdef USE_PTHREADS
#ifdef HAVE_PTHREAD_H
# include <pthread.h>
#endif
#endif

#include "log.h"
#include "task-maker.h"

#ifdef ENABLE_CRYPTO_LAYER
#include <peks/peks.h>
#endif

/* ---------------------------------------------------------------------- *
 *                  definitions and data structures                       *
 * ---------------------------------------------------------------------- */

/* The XMALLOC() is assumed to always (not NULL) return zero
   initialized memory */
#define XMALLOC(x)     emalloc (x)
#define XFREE(x)       efree (&(x))

/* XREALLOC() always returns -- possible uninitialized -- memory */
#define WANT_TASKREALLOC
#define XREALLOC(x,y)  taskrealloc (x,y) /* defined below */

/* to be called XERROR (("problem", ..)) */
#define XERROR(msg)  {int e = errno; log msg; errno = e;}

/* the minimal size of the task array, this avoids unnecessary
   reallocation of small data */
#define MIN_TASKLIST  5

/* message payload structures */
typedef
struct _start_new_task {
  int   socket;
  void (*fn) (taskdef*, void*) ;
  void  *fn_arg ;
} start_new_task ;

#ifndef USE_PTHREADS
#define pthread_mutex_lock(x)
#define pthread_mutex_unlock(x)
#define pthread_mutex_destroy(x)
#define pthread_mutex_init(x,y)
#endif /* USE_PTHREADS */

/* ---------------------------------------------------------------------- *
 *                  private data and structures                           *
 * ---------------------------------------------------------------------- */

/* ingnore signals template */
static struct sigaction ign_handler = {SIG_IGN} ;

/* ---------------------------------------------------------------------- *
 *                   private functions: helpers                           *
 * ---------------------------------------------------------------------- */

#ifdef WANT_TASKREALLOC
static void *
taskrealloc 
  (void  *p, 
   size_t n)
{
  if ((p =  realloc (p, n)) == 0) {
    XERROR (("Cannot reallocate %u bytes -- aborting\n", n)) ;
    exit (1) ;
  }
  return p ;
}
#endif

static void
set_wait_for_some_time_flag
  (taskdef *t)
{
  pthread_mutex_lock (&t->mx);
  /* give the pthreads a chance to signal an alarm */
  t->waiting_for_some_time = 1 ;
  pthread_mutex_unlock (&t->mx);
}

static void
res_wait_for_some_time_flag
  (taskdef *t)
{
  pthread_mutex_lock (&t->mx);
  /* give the pthreads a chance to signal an alarm */
  t->waiting_for_some_time = 0 ;
  pthread_mutex_unlock (&t->mx);
}

static void
send_selectable_event
  (taskdef *t)
{
  pthread_mutex_lock (&t->mx);
  if (t->waiting_for_some_time != 0) {

    /* save alarm structure */
    int tmo_save = alarm (0) ;
    struct sigaction save_handler;
    sigaction (SIGALRM, &ign_handler, &save_handler);

    /* signal to the master process */
    raise (SIGALRM);

    /* restore alarm structure */
    sigaction (SIGALRM, &save_handler, 0);
    alarm (tmo_save) ;
  }
  pthread_mutex_unlock (&t->mx);
}

/* ---------------------------------------------------------------------- *
 *                   private functions: io handling                       *
 * ---------------------------------------------------------------------- */

static void
add_read_selector 
  (taskdef *me,
   int      fd)
{
# ifdef HAVE_POLL
  int n = me->dim_read_poller ;
  me->read_poller = (me->dim_read_poller ++ == 0)
    ? XMALLOC                   (me->dim_read_poller * sizeof (struct pollfd))
    : XREALLOC (me->read_poller, me->dim_read_poller * sizeof (struct pollfd))
    ;
  me->read_poller [n].fd     =     fd ;
  me->read_poller [n].events = POLLIN ;
# endif

# ifdef HAVE_SELECT
  if (me->limit_read_selector == 0)
    FD_ZERO (&me->read_selector);
  FD_SET (fd, &me->read_selector) ;
  if (fd > me->limit_read_selector)
    me->limit_read_selector = fd + 1;
# endif
}

static void
del_read_selector
  (taskdef *me,
   int      fd)
{
  unsigned n ;

# ifdef HAVE_POLL
  n = me->dim_read_poller;
  while (n --) 
    if (me->read_poller [n].fd == fd) {
      if ( -- me->dim_read_poller == 0) {
	XFREE (me->read_poller);
	break ;
      }
#     ifdef HAVE_MEMMOVE
      memmove (me->read_poller + n, me->read_poller + n + 1, 
	       (me->dim_read_poller - n) * sizeof (struct pollfd));
#     else
      while (n ++ < me->dim_read_poller)
	me->read_poller [n - 1] = me->read_poller [n] ;
#     endif /* HAVE_MEMMOVE */
      me->read_poller =	XREALLOC
	(me->read_poller, me->dim_read_poller * sizeof (struct pollfd));
      break ;
    }
# endif

# ifdef HAVE_SELECT
  FD_CLR (fd, &me->read_selector) ;
  n = me->limit_read_selector ;
  while (n -- && !FD_ISSET (n, &me->read_selector))
    ;
  me->limit_read_selector = n + 1;
# endif
}

static void
append_message_queue
  (mesg **Q,
   mesg  *m)
{
  while (*Q != 0) 
    Q = &(*Q)->next ;
  *Q = m ;
}


/* ---------------------------------------------------------------------- *
 *                   private functions: con/destructor                    *
 * ---------------------------------------------------------------------- */

/* find destination record */
static int
find_task_inx
  (taskdef *me,
   int     tid)
{
  if (me == 0)
    return -1;
  
  if (me->sparse_list) {
    int n = me->dim_tasks ;
    while (n --)
      if (me->tasks [n].taskid == tid)
	return n;
    return -1;
  }
  if (tid < me->dim_tasks && me->tasks [tid].taskid == tid)
    return tid ;
  return -1 ;
}


static void
destroy_taskrec
  (taskdef  *me,
   unsigned tid)
{
  taskrec* t;
  int n ;
  
  /* get the task record */
  if (me == 0 || (n = find_task_inx (me, tid)) < 0)
    return ;
  t = me->tasks + n ;
  
  /* close the communication channels */
  switch (t->task_type) {
    taskdef *d;
  case ISA_SHMEM_THREAD_MASTER :
    /* reset the master reference record on the slave */
    if ((d = t->io.slave.task) == 0)
      break ;
    d->tasks [n].task_type = IS_NOTHING ;
    destroy_taskrec (d, n) ; /* recursion ! */
    break ;
  case ISA_SHMEM_THREAD_SLAVE :
    /* reset the slave reference record on the master */
    if ((d = t->io.master.task) == 0)
      break ;
    d->tasks [n].task_type = IS_NOTHING ;
    destroy_taskrec (d, n) ; /* recursion ! */
    break ;
  case ISA_PIPE_CHILD_MASTER :
  case ISA_PIPE_CHILD_SLAVE :
  case ISA_PIPE_THREAD_MASTER :
    close (t->io.pipe.send) ;
    close (t->io.pipe.recv) ;
    del_read_selector (me, t->io.pipe.recv);
    break ;
  } /* switch */
 
  /* do not delete the task array, completely */
  if ((n = me->dim_tasks) <= MIN_TASKLIST) {
    memset (t, 0, sizeof (taskrec));
    return;
  }
  
  /* adjust the dimension to the first entry that is non-empty */
  while ((tid = me->dim_tasks) >= MIN_TASKLIST &&
	 me->tasks [tid - 1].taskid == 0)
    -- me->dim_tasks ;
  
  /* shrink the task list */
  if (n > me->dim_tasks)
    me->tasks = XREALLOC (me->tasks, me->dim_tasks * sizeof (taskrec));
}



static void
destroy_taskrec_all
  (taskdef *me)
{
  int n ;
  
  if (me == 0)
    return ;

  for (n = me->dim_tasks-1; n; n --)
    if (me->tasks [n].taskid)
      /* close the communication channels */
      switch (me->tasks [n].task_type) {
	taskdef *d;
      case ISA_SHMEM_THREAD_MASTER :
	/* reset the master reference record on the slave */
	if ((d = me->tasks [n].io.slave.task) == 0)
	  break ;
	d->tasks [n].task_type = IS_NOTHING ;
	destroy_taskrec (d, n);
	break ;
      case ISA_SHMEM_THREAD_SLAVE :
	/* reset the slave reference record on the master */
	if ((d = me->tasks [n].io.master.task) == 0)
	  break ;
	d->tasks [n].task_type = IS_NOTHING ;
	destroy_taskrec (d, n);
	break ;
      case ISA_PIPE_CHILD_MASTER :
      case ISA_PIPE_CHILD_SLAVE :
      case ISA_PIPE_THREAD_MASTER :
	close (me->tasks [n].io.pipe.send) ;
	close (me->tasks [n].io.pipe.recv) ;
	break ;
      } /* switch */

  /* clean up pending messages */
  while (me->incoming != 0) {
    mesg *m = me->incoming->next ;
    XFREE (me->incoming) ;
    me->incoming = m ;
  }

  pthread_mutex_destroy (&me->mx) ;
# ifdef HAVE_POLL
  if (me->read_poller != 0)
    XFREE (read_poller) ;
# endif
  XFREE (me->tasks) ;
  XFREE (me) ;
}


  
static taskdef*
activate_taskrec_slot
  (taskdef *me)
{
  int n ;

  if (me == 0) {
    /* formally activates slot zero, which is a special case as
       in the normal context, 0 means: no activation at all */
    me = XMALLOC (sizeof (taskdef));
    me->    tasks = XMALLOC (MIN_TASKLIST * sizeof (taskrec));
    me->dim_tasks =          MIN_TASKLIST ;
    pthread_mutex_init (&me->mx, 0);
    return me ;
  }

  /* find the index n > 0 of a a free slot */
  for (n = 1; n < me->dim_tasks; n ++)
    if (me->tasks [n].taskid == 0) 
      goto ready_to_go ;
  
  /* append a new, clean task record at the end */
  me->tasks = XREALLOC (me->tasks, (++ me->dim_tasks) * sizeof (taskrec));
  memset (me->tasks + n, 0, sizeof (taskrec));

 ready_to_go: /* n in the index of a free entry */
  me->tasks [n].taskid = 
    me->act_task = (me->sparse_list ? ++ me->sparse_list : n) ;
  return me ;
}


static taskdef*
activate_taskrec_type
  (taskdef  *me,
   taskobj type)
{
  int n, created_me = (me == 0) ;
  
  /* activate_taskrec_slot (0)->act_task == 0 */
  me = activate_taskrec_slot (me);
  if ((n = me->act_task) == 0) 
    me->tasks [1].taskid = me->act_task = n = 1 ;
  
  /* create the io stuff */
  switch (me->tasks [n].task_type = type) {
    int fd [2] ;
  case ISA_PIPE_CHILD_MASTER :
  case ISA_PIPE_THREAD_MASTER :
    if (pipe (fd) < 0)
      break ;
    me->tasks [n].io.pipe.send = fd [1] ;
    me->tasks [n].io.pipe.recv = fd [0] ;
    add_read_selector (me, fd [0]) ;
    return me ;
  case ISA_SHMEM_THREAD_MASTER :
  case ISA_SHMEM_THREAD_SLAVE :
  case ISA_PIPE_CHILD_SLAVE :
  case ISA_PIPE_THREAD_SLAVE :
  case ISA_PIPE_ROUTER_SLAVE :
    return me ;
  }
  
  /* error, undo the last creation step */
  if (created_me == 0) {
    memset (me->tasks + n, 0, sizeof (taskrec));
    return 0;
  }

  /* destroy the descriptor currently created */
  pthread_mutex_destroy (&me->mx) ;
  XFREE (me->tasks);
  XFREE (me);
  return 0;
}

/* ---------------------------------------------------------------------- *
 *               private functions: start a master tasks                  *
 * ---------------------------------------------------------------------- */

static int
fork_a_child
  (taskdef *me,
   int newtype)
{    
  taskdef *u;
  
  /* get current task record */
  int      n  = me->act_task ;
  taskrec *t  = me->tasks + n ;
  
# ifdef ENABLE_CRYPTO_LAYER
  if (t->soc >= -1)
    /* when EBADF is set, there was no data stream assigned (no problem) */
    if ((t->tid = io_ctrl 
	 (t->soc, IO_REGISTER_THREAD, 0, 1)) < 0 && errno != EBADF) {
      XERROR (("Cannot register on peks client: %s", peks_strerr (errno))) ;
      return -1;
    }
# endif /* ENABLE_CRYPTO_LAYER */
  
  if ((t->tt.child.pid = fork ()) == 0) {

    unsigned loop_it = 1 ;                /* DEBUG DEBUG DEBUG DEBUG */
    while (loop_it) {                     /* DEBUG DEBUG DEBUG DEBUG */
      fprintf                             /* DEBUG DEBUG DEBUG DEBUG */
	(stderr, "catch me, pid=%d\n", getpid());     /* DEBUG DEBUG */
      sleep (10);                         /* DEBUG DEBUG DEBUG DEBUG */
    }                                     /* DEBUG DEBUG DEBUG DEBUG */
    
    /* child: set the argument record as new master entry */
    u          = memcpy (XMALLOC (sizeof (taskdef)), me, sizeof (taskdef));
    u->tasks   = memcpy (XMALLOC (sizeof (taskrec)),  t, sizeof (taskrec));
    u->tasks->task_type = newtype ;
    add_read_selector (u, t->io.pipe.recv) ;

    /* clean up unnecessary things: destroy all but the t entry */
    t->taskid = 0;
    destroy_taskrec_all (me);

    /* me->tasks [id] == id is not true, anymore */
    u->sparse_list = 1 ; 
    
#   ifdef ENABLE_CRYPTO_LAYER
    if (u->tasks->soc >= -1) {
      io_ctrl (u->tasks->soc, IO_ACTIVATE_THREAD, &u->tasks->tid, 1);
      io_ctrl (u->tasks->soc, IO_SYNTHETIC_PID,                0, 1);
    }
#   endif /* ENABLE_CRYPTO_LAYER */

    /* execute that function and die */
    (*u->tasks->fn)(u, u->tasks->fn_arg);
  
#   ifdef ENABLE_CRYPTO_LAYER
    io_ctrl (u->tasks->soc, IO_DESTROY_THREAD,  &u->tasks->tid, 1);
#   endif
    
    exit (0);
    /*@NOTREACHED@*/
  }

# ifdef ENABLE_CRYPTO_LAYER
  /* the parent does not want to keep this thread */
  io_ctrl (t->soc, IO_UNLINK_THREAD, &t->tid, 1);
# endif

  if (t->tt.child.pid > 0) 
    return n ;
  XERROR (("Could not fork (%s)\n", strerror (errno)));
  return -1 ;
}

static void* 
fn_wrapper
  (void *u) 
{
  (*((taskdef*)u)->tasks->fn)((taskdef*)u, ((taskdef*)u)->tasks->fn_arg);
  return 0;
}

static int
start_a_thread
  (taskdef *me,
   int newtype)
{ 
  taskdef *u ;

  /* get current task record */
  int      n = me->act_task ;
  taskrec *t = me->tasks + n ;
  
  /* make new task descriptor */
  u          = memcpy (XMALLOC (sizeof (taskdef)), me, sizeof (taskdef));
  u->tasks   = memcpy (XMALLOC (sizeof (taskrec)),  t, sizeof (taskrec));
  u->tasks->task_type = newtype ;

  /* me->tasks [id] == id is not true, anymore */
  u->sparse_list = 1 ; 

  switch (newtype) {
  case ISA_SHMEM_THREAD_SLAVE :
    /* set up communication */
    u->tasks->io.master.task = me ;
    me->tasks->io.slave.task =  u ;
    break;
  case ISA_PIPE_THREAD_SLAVE :
    u->tasks->io = t->io ;
  }

# ifdef USE_PTHREADS 
  if (pthread_create (&t->tt.thread.tid, 0, fn_wrapper, me) == 0)
    return n ;
  XERROR (("Could not create thread (%s).\n", strerror (n))) ;
  return -1 ;
# else
  fn_wrapper (me);
  return n ;
# endif
}


static int
create_master_task
  (taskdef  *me,
   taskobj what,
   int   socket,
   void (*fn) (taskdef*, void*),
   void *fn_arg)
{
  taskrec *t;
  int n, nn, tid, master;
  int (*starter) (taskdef *, int);
  
  if (me == 0) {
    XERROR (("Argument error: No master descriptor.\n"));
    return -1;
  }
  if (fn == 0) {
    XERROR (("Argument error: Cannot fork or thread (*0)().\n"));
    return -1;
  }
  
  switch (what) {

    /* FORK, PIPE */
# ifndef USE_PTHREADS 
  case START_A_THREAD :
    /* fall back to fork unless pthreads, available */
# endif
  case FORK_A_CHILD :
    what = ISA_PIPE_CHILD_SLAVE ;
  case ISA_PIPE_CHILD_SLAVE : 
    starter = fork_a_child ;
    master  = ISA_PIPE_CHILD_MASTER;
    break;

    /* PTHREAD, PIPE */
  case ISA_PIPE_THREAD_SLAVE :
    starter = start_a_thread ;
    master = ISA_PIPE_THREAD_MASTER;
    break;

    /* PTHREAD, SHARED-MEMORY */
# ifdef USE_PTHREADS 
  case START_A_THREAD : /* use pthreads, if possible */
    what = ISA_SHMEM_THREAD_SLAVE ;
# endif
  case ISA_SHMEM_THREAD_SLAVE :
    starter = start_a_thread ;
    master = ISA_SHMEM_THREAD_MASTER;
    break;

  default:
    XERROR (("Illegal communication type %u.\n", what)) ;
    return -1;
  }

  /* activate task record */
  if ((me = activate_taskrec_type (me, master)) == 0)
    return -1 ;
  
  t->fn     = fn ;
  t->fn_arg = fn_arg ;
  t->soc    = socket;

  return (*starter) (me, what) ;
}

/* ---------------------------------------------------------------------- *
 *                   private functions: messages                          *
 * ---------------------------------------------------------------------- */

static int
send_message
  (taskdef    *me,
   int      trgid,
   taskobj  mtype,
   unsigned ackid,
   char     *data,
   unsigned   len)
{
  mesg *m ;
  int size, n, mid ;
  taskrec *t ;

  if (me == 0 || (len != 0 && data == 0) ||
      (n = find_task_inx (me, trgid)) < 0)
    return -1 ;

  t = me->tasks + n ;

  if (len == 0 && data != 0)
    len = strlen (data) + 1;
    
  /* create message body */
  m = XMALLOC (size = sizeof (mesg) - 1 + len);
  m->size      = size ;
  m->src       = me->tasks->taskid ;
  m->trg       = trgid ;
  m->mesg_type = mtype ;
  m->ackid     = ackid ;
  if ((mid = ++ t->msgid) == 0)
    mid = ++ t->msgid ;
  m->msgid     = mid ;

  if (len) 
    memcpy (m->data, data, len);

  switch (t->task_type) {
    /* With shared memory, messages are placed immediately
       into the receiver queue. As this is shared, we must
       take care to assure mutual exlusive access */
  case ISA_SHMEM_THREAD_MASTER :
    pthread_mutex_lock (t->io.slave.task->mx) ;
    append_message_queue (&t->io.slave.task->incoming, m) ;
    t->io.slave.task->shared_count ++ ; /* set flag */
    pthread_mutex_lock (t->io.slave.task->mx) ;
    send_selectable_event (me);
    return mid ;
  case ISA_SHMEM_THREAD_SLAVE :
    pthread_mutex_lock (t->io.master.task->mx) ;
    append_message_queue (&t->io.master.task->incoming, m) ;
    t->io.master.task->shared_count ++ ; /* set flag */
    pthread_mutex_lock (t->io.master.task->mx) ;
    send_selectable_event (me);
    return t->msgid ;
    /* send tata to wherever they need to go to */
  case ISA_PIPE_THREAD_MASTER :
  case ISA_PIPE_CHILD_MASTER :
  case ISA_PIPE_THREAD_SLAVE :
  case ISA_PIPE_CHILD_SLAVE :
    n = write (t->io.pipe.send, m, size) < 0 ? -1 : mid ;
    XFREE (m);
    return n ;
  case ISA_PIPE_ROUTER_SLAVE :
    t = me->tasks ;
    switch (t->task_type) {
    case ISA_PIPE_THREAD_MASTER :
    case ISA_PIPE_CHILD_MASTER :
    case ISA_PIPE_THREAD_SLAVE :
    case ISA_PIPE_CHILD_SLAVE :
      n = write (t->io.pipe.send, m, size) < 0 ? -1 : mid ;
      XFREE (m);
      return n ;
    }
  }
  XFREE (m);
  return -1 ;
}


static void
recv_single_message_from_io
  (mesg **M,
   int   fd)
{
  mesg r, *in;
  /* read the message header */
  if (read (fd, &r, sizeof (r)-1) < 0)
    return ;
  in = memcpy (XMALLOC (sizeof (r) - 1 + r.size), &r, sizeof (r)-1);
  /* are there message data, at all ? */
  if (r.size < sizeof (r))
    return ;
  /* read the message data */
  if (read (fd, &in->data, 1 + r.size-sizeof (r)) < 0) {
    XFREE (in);
    return ;
  }
  append_message_queue (M, in);
}


static void
recv_messages
  (taskdef *me,
   int  tmo_ms)
{
#ifdef HAVE_GETTIMEOFDAY2
  /* current time in millisecs, or centisecs etc. */
  struct timeval start, tv ;
  gettimeofday2 (&start, 0);
#else
  time_t tv, start = time (0);
#endif

  int n ;
  
  /* reset flags */
  pthread_mutex_lock (me->mx) ;
  me->shared_count = 0;
  pthread_mutex_unlock (me->mx);

# ifdef HAVE_SELECT
  if (me->limit_read_selector != 0) {
    fd_set rfds = me->read_selector ;
    struct timeval tv, *T = 0;
    if (tmo_ms >= 0) {
      tv.tv_usec = (tmo_ms % 1000) * 1000 ;
      tv.tv_sec  =  tmo_ms / 1000;
      T = &tv ;
    }
    if (tmo_ms) set_wait_for_some_time_flag (me) ;
    n = select (me->limit_read_selector, &rfds, 0, 0, T) ;
    if (tmo_ms) res_wait_for_some_time_flag (me) ;
    if (n < 0 && n != EINTR)
      XERROR (("Selecting error in recv_messages (%s)\n", strerror (errno))) ;
    if (n > 0) {
      /* get all messages, available */
      n = me->limit_read_selector ;
      while (n --)
	if (FD_ISSET (n, &rfds)) {
	  /* as some thread may have shared memory access to this
	     queue, we need mutual excluseive access */
	  pthread_mutex_lock (me->mx) ;
	  recv_single_message_from_io (&me->incoming, n);
	  pthread_mutex_unlock (me->mx);
	}
    }
  }
# else
# ifdef HAVE_POLL
  if (me->dim_read_poller != 0) {
    if (tmo_ms) set_wait_for_some_time_flag (me) ;
    n = poll (me->read_poller, me->dim_read_poller, tmo_ms) ;
    if (tmo_ms) res_wait_for_some_time_flag (me) ;
    if (n < 0 && n != EINTR)
      XERROR (("Polling error in recv_messages (%s).\n", strerror (errno))) ;
    if (n > 0) {
      /* get all messages, available */
      n = me->dim_read_poller ;
      while (n --) {
	if (me->read_poller [n].revents & POLLIN) {
	  /* as some thread may have shared memory access to this
	     queue, we need mutual excluseive access */
	  pthread_mutex_lock (me->mx) ;
	  recv_single_message_from_io (&me->incoming, n);
	  pthread_mutex_unlock (me->mx);
	}
	me->read_poller [n].revents = 0 ;
      }
    }
  }
# endif
# endif
  
  /* check flags whether we got some shared memory messages, already */
  pthread_mutex_lock (me->mx) ;
  n = me->shared_count;
  pthread_mutex_unlock (me->mx);
  if (n) return ;

  /* calculate the time elapsed, already */
  if (tmo_ms) {
#   ifdef HAVE_GETTIMEOFDAY2
    gettimeofday2 (&tv, 0);
    if ((tv.tv_usec -= start.tv_usec) < 0) {
      tv.tv_usec += 1000000;
      tv.tv_sec  -=       1;
    }
    /* this happens only once in a life time */
    if ((tv.tv_sec -= start.tv_sec) < 0 ||
	(tmo_ms -= ((tv.tv_sec*1000) + ((tv.tv_usec + 999)/1000))) < 0)
      return ;
#   else
    tv = time (0);
    if ((tv -= start + 1) <= 0)
      tv = 0 ;
    tv *= 1000 ;
    tv +=  500 ;
    if ((tmo_ms -= tv) <= 0)
      tmo_ms = 1 ;
#   endif
  }
  
  /* wait to get some messages */
# ifdef HAVE_SELECT
  {
    struct timeval tv;
    tv.tv_sec  =  0;
    tv.tv_usec = tmo_ms;
    select (FILE_LOCK_POLL_FD, 0, 0, 0, &tv);
  }
# else
# ifdef HAVE_POLL
  {
    struct pollfd pfd;
    memset (&pfd, 0, sizeof (pfd));
    pfd.fd = FILE_LOCK_POLL_FD;
    poll (&pfd, 1, tmo_ms);
  }
# else
  sleep ((tmo_ms + 999)/1000);
# endif
# endif
  
}

/* ---------------------------------------------------------------------- *
 *                            public functions                            *
 * ---------------------------------------------------------------------- */

taskdef *
init_tasks
  (void)
{
  return activate_taskrec_slot (0);
}

int
new_task
  (taskdef *me,
   taskobj what,
   int   tmo_ms,
   int   socket,
   void (*fn) (taskdef*, void*),
   void *fn_arg)
{
  start_new_task payl;
  mesg **M, *m;
  int n, inx ;
  
  if (me == 0)
    return -1;
  
  /* are we master task ? */
  if (me->tasks->taskid == 0)
    return create_master_task (me, what, socket, fn, fn_arg) == 0
      ? -1 
      : me->act_task 
      ;

  /* tell the master to start the new task */
  payl.socket = socket ;
  payl.fn     = fn ;
  payl.fn_arg = fn_arg ;

  /* send start message, get answer */
  if ((n = send_message 
       (me, me->tasks [n].taskid, what, 0, (char*)&payl, sizeof (payl))) >= 0)
    recv_messages (me, tmo_ms) ;

  /* Check, whether we got the anser n. As some thread may have
     shared memory access to this queue, we need mutual excluseive
     access to the imcoming queue. */
  pthread_mutex_lock (me->mx) ;
  for (M = &me->incoming; (m = *M) != 0; M = &m->next) {
    if (m->ackid != n) 
      continue ;
    /* we found the answer, unlink that message */
    *M = m->next ;
    /* so it is save to unkock, now */
    pthread_mutex_unlock (me->mx);
    switch (m->mesg_type) {
    case MSG_THREADID : 
      n = (unsigned)(m->data); /* this is the task id to be returned */
      activate_taskrec_type (me, ISA_PIPE_ROUTER_SLAVE);
      break ;
    case MSG_ACK :                                  n =  0; break;
    case MSG_ERRNO :  errno = (unsigned)(m->data);  n = -1; break;
    default:          errno = 0 ;                   n = -1;
    }
    XFREE (m);
    return n;
  }
  pthread_mutex_unlock (me->mx);

  errno = 0 ;
  return -1 ;
}


mesg *
get_message
  (taskdef *me,
   int  tmo_ms)
{
  unsigned tid, type ;
  mesg *hold ;
  int once = 0 ;

  for (;;) {
    mesg *m, **M ;
    
    /* As some thread may have shared memory access to this queue, we
       need mutual excluseive access to the imcoming queue. */
    pthread_mutex_lock (me->mx) ;
    M = &me->incoming ;
    while (m = *M, m != 0) {
      /* is there something to do ? */
      switch (m->mesg_type) {
      case START_A_THREAD :
      case FORK_A_CHILD :
      case ISA_PIPE_CHILD_SLAVE : 
      case ISA_PIPE_THREAD_SLAVE :
      case ISA_SHMEM_THREAD_SLAVE :
	/* unlink that message and process it */
	*M = m->next ;
	/* so it is save to unkock, now */
	pthread_mutex_unlock (me->mx);
	tid = create_master_task 
	  (me, m->mesg_type, 
	   ((start_new_task*)(m->data))->socket,
	   ((start_new_task*)(m->data))->fn,
	   ((start_new_task*)(m->data))->fn_arg) ;
	if (tid >= 0) 
	  type = MSG_THREADID ;
	else {
	  tid = errno ; 	  /* send error message */
	  type = MSG_ERRNO ;
	} 
	send_message (me, m->src, type, m->msgid, (char*)&tid, sizeof (tid));
	XFREE (m);
	/* re-lock, again */
	pthread_mutex_lock (me->mx) ;
	continue ;
      }
      /* loop through all messages */
      M = &m->next ;
    }
    /* the incoming queue is still locked */
    if (me->incoming != 0) {
      /* keep processed messages on an extra queue */
      append_message_queue (&hold, me->incoming) ;
      me->incoming = 0 ;
    }
    pthread_mutex_unlock (me->mx);
    
    /* recv_messages() is called once */
    if (once ++) break ;
    recv_messages (me, tmo_ms);
  }

  pthread_mutex_lock (me->mx) ;
  if ((me->incoming = hold) != 0) {
    me->incoming = hold->next ;
    hold->next   =         0  ;
  }
  pthread_mutex_unlock (me->mx);

  return hold ;
}

/* ---------------------------------------------------------------------- *
 *        test                                                            *
 * ---------------------------------------------------------------------- */

static void
taskfn 
 (taskdef* me, 
  void* info)
{
  unsigned stay_in_taskfn = 1;

  while (stay_in_taskfn) {
    fprintf (stderr, "killme, pid = %d", getpid ());
    sleep (5);
  }
}


task_loop 
  (void)
{
  unsigned stay_in_task_loop = 1;
  taskdef *task = init_tasks ();

  
  while (stay_in_task_loop) {
    
    int task_id =  new_task (task, FORK_A_CHILD, 0, -1, taskfn, 0);
    sleep (10) ;
  }
}
